import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_none


@pytest.mark.parametrize("value", [
    "none",
    "None",
    "null",
    "NULL",
    "Nil",
])
def test_parse_none(value):
    assert parse_none(value) is None


@pytest.mark.parametrize("value", [
    "",
    "0",
    "undefined",
])
def test_parse_invalid_none(value):
    with raises(Exception) as exc_info:
        parse_none(value)

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == f"Failed to parse {value!r} as None"



import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_bool


@pytest.mark.parametrize("value", [
    "y",
    "yes",
    "t",
    "true",
    "on",
    "1",
])
def test_parse_truthy_bool(value):
    assert parse_bool(value) is True


@pytest.mark.parametrize("value", [
    "n",
    "no",
    "f",
    "false",
    "off",
    "0",
])
def test_parse_falsy_bool(value):
    assert parse_bool(value) is False


@pytest.mark.parametrize("value", [
    "",
    "42",
    "banana",
])
def test_parse_invalid_bool(value):
    with raises(Exception) as exc_info:
        parse_bool(value)

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == f"Failed to parse {value!r} as bool"



import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_str


@pytest.mark.parametrize("value", [
    "banana",
    " banana",
    "banana ",
    " banana ",
    "  banana  ",
])
def test_parse_str(value):
    assert parse_str(value) == "banana"


def test_parse_empty_str():
    with raises(Exception) as exc_info:
        parse_str("")

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == "Failed to parse '' as non-empty str"



import os
from functools import partial
from typing import Any, Callable, Dict, Mapping, Tuple, Union, cast

from niltype import Nil, NilType

from ._future_value import FutureValue, ValueType
from .errors import EnvKeyError
from .parsers import (
    parse_as_is,
    parse_bool,
    parse_float,
    parse_int,
    parse_none,
    parse_str,
    parse_tuple,
)


class Environment:
    def __init__(self, environ: Mapping[str, str] = os.environ, *, prefix: str = "") -> None:
        self._environ = environ
        self._prefix = prefix

    def __repr__(self) -> str:
        return f"cabina.Environment({self._environ!r})"

    def get(self, name: str, default: Union[NilType, ValueType] = Nil,
            parser: Callable[[str], ValueType] = parse_as_is) -> ValueType:
        if self._prefix:
            name = self._prefix + name
        try:
            value = self._environ[name]
        except KeyError:
            if default is Nil:
                raise EnvKeyError(f"{name!r} does not exist") from None
            return default
        else:
            return parser(value)

    def raw(self, name: str, default: Union[NilType, ValueType] = Nil,
            parser: Callable[[str], ValueType] = parse_as_is) -> ValueType:
        kwargs: Dict[str, Any] = {}
        if default is not Nil:
            kwargs["default"] = default
        if parser is not parse_as_is:
            kwargs["parser"] = parser
        return cast(ValueType, FutureValue[ValueType](self.get, name, **kwargs))

    def __call__(self, name: str, default: Union[NilType, ValueType] = Nil,
                 parser: Callable[[str], ValueType] = parse_as_is) -> ValueType:
        return self.raw(name, default, parser)

    def none(self, name: str, default: Union[NilType, None] = Nil) -> None:
        assert isinstance(default, (type(None), NilType))
        return self(name, default, parse_none)

    def bool(self, name: str, default: Union[NilType, bool] = Nil) -> bool:
        assert default is Nil or isinstance(default, bool)
        return self(name, default, parse_bool)

    def int(self, name: str, default: Union[NilType, int] = Nil) -> int:
        assert default is Nil or isinstance(default, int)
        return self(name, default, parse_int)

    def float(self, name: str, default: Union[NilType, float] = Nil) -> float:
        assert default is Nil or isinstance(default, float)
        return self(name, default, parse_float)

    def tuple(self, name: str,
              default: Union[NilType, Tuple[ValueType, ...]] = Nil, *,
              separator: str = ",",
              subparser: Callable[[str], Any] = parse_str) -> Tuple[ValueType, ...]:
        assert default is Nil or isinstance(default, tuple)
        parser = partial(parse_tuple, separator=separator, subparser=subparser)
        return self(name, default, parser)

    def str(self, name: str, default: Union[NilType, str] = Nil) -> str:
        assert default is Nil or isinstance(default, str)
        return self.raw(name, default, parse_str)



__all__ = ("ConfigError", "ConfigAttrError", "ConfigKeyError",
           "EnvError", "EnvKeyError", "EnvParseError", "ConfigEnvError",
           "Error",)


class Error(Exception):
    pass


class ConfigError(TypeError, Error):
    pass


class ConfigAttrError(AttributeError, ConfigError):
    pass


class ConfigKeyError(KeyError, ConfigError):
    def __str__(self) -> str:
        return ConfigError.__str__(self)


class EnvError(Error):
    pass


class EnvKeyError(KeyError, EnvError):
    def __str__(self) -> str:
        return EnvError.__str__(self)


class EnvParseError(TypeError, EnvError):
    pass


class ConfigEnvError(ValueError, EnvError, ConfigError):
    pass



from typing import Any, Callable, Generic, TypeVar, Union

from niltype import Nil, NilType

ValueType = TypeVar("ValueType")


class FutureValue(Generic[ValueType]):
    def __init__(self, accessor: Callable[..., ValueType], *args: Any, **kwargs: Any) -> None:
        self._accessor = accessor
        self._args = args
        self._kwargs = kwargs
        self._value: Union[ValueType, NilType] = Nil

    def fetch(self) -> ValueType:
        self._value = self._accessor(*self._args, **self._kwargs)
        return self._value

    def get(self) -> ValueType:
        if self._value is Nil:
            return self.fetch()
        return self._value

    def __repr__(self) -> str:
        args = [repr(arg) for arg in self._args]
        str_args = ", ".join(args)

        kwargs = [f"{key}={val!r}" for key, val in self._kwargs.items()]
        str_kwargs = ", ".join(kwargs)

        if (len(args) > 0) and (len(kwargs) > 0):
            return f"FutureValue({str_args}, {str_kwargs})"
        elif len(args) > 0:
            return f"FutureValue({str_args})"
        return f"FutureValue({str_kwargs})"



from typing import Any, Callable, Tuple

from ..errors import EnvParseError

__all__ = ("parse_as_is", "parse_none", "parse_bool", "parse_int",
           "parse_float", "parse_str", "parse_tuple",)


def parse_as_is(value: Any) -> Any:
    return value


def parse_none(value: str) -> None:
    if value.lower() in ("none", "null", "nil",):
        return None
    raise EnvParseError(f"Failed to parse {value!r} as None")


def parse_bool(value: str) -> bool:
    if value.lower() in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif value.lower() in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise EnvParseError(f"Failed to parse {value!r} as bool")


def parse_int(value: str, *, base: int = 10) -> int:
    try:
        return int(value, base)
    except ValueError:
        if base == 10:
            raise EnvParseError(f"Failed to parse {value!r} as int") from None
        raise EnvParseError(f"Failed to parse {value!r} as int with base {base}") from None


def parse_float(value: str) -> float:
    try:
        return float(value)
    except ValueError:
        raise EnvParseError(f"Failed to parse {value!r} as float") from None


def parse_str(value: str, *,
              not_empty: bool = True,
              trim: Callable[[str], str] = str.strip) -> str:
    parsed = trim(value)
    if not_empty and parsed == "":
        raise EnvParseError(f"Failed to parse {value!r} as non-empty str")
    return parsed


def parse_tuple(value: str, *, separator: str = ",",
                subparser: Callable[[str], Any] = parse_str) -> Tuple[Any, ...]:
    if value.startswith("(") and value.endswith(")"):
        value = value[1:-1]
    if value.endswith(separator):
        value = value[:-len(separator)]
    parsed = value.split(separator)
    try:
        return tuple(subparser(x) for x in parsed)
    except EnvParseError as e:
        raise EnvParseError(f"Failed to parse {value!r} as tuple: {e}") from None



import inspect
import sys
from typing import (
    Any,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Optional,
    Tuple,
    Union,
    ValuesView,
)

from niltype import Nil, NilType

from ._future_value import FutureValue
from .errors import (
    ConfigAttrError,
    ConfigEnvError,
    ConfigError,
    ConfigKeyError,
    EnvKeyError,
    EnvParseError,
)

_Section = None
_Config = None


def _is_dunder(name: str) -> bool:
    return (name[:2] == name[-2:] == '__' and
            name[2:3] != '_' and
            name[-3:-2] != '_' and
            len(name) > 4)


def _is_section(cls: Any) -> bool:
    return _Section is not None and issubclass(cls, _Section)


def _is_config(cls: Any) -> bool:
    return _Config is not None and issubclass(cls, _Config)


def _is_subclass(cls: Any, cls_type: Any) -> bool:
    return inspect.isclass(cls) and issubclass(cls, cls_type)


class UniqueDict(Dict[str, Any]):
    def __init__(self, namespace: str) -> None:
        super().__init__()
        self.__namespace = namespace

    def __setitem__(self, key: str, value: Any) -> None:
        if not _is_dunder(key) and key in self:
            raise ConfigError(f"Attempted to reuse {key!r} in {self.__namespace!r}")
        super().__setitem__(key, value)


class MetaBase(type):
    @classmethod
    def __prepare__(mcs, name: str, bases: Tuple[Any]) -> UniqueDict:  # type: ignore
        return UniqueDict(name)

    def __init__(cls, name: str, bases: Tuple[Any], attrs: Dict[str, Any]) -> None:
        super().__init__(name, bases, attrs)

        for base in bases:
            if not _is_subclass(base, (_Config, _Section)):
                raise ConfigError(f"Attempted to inherit {base!r}")

        if _is_config(cls) or _is_section(cls):
            cls.__frozen__ = False
            cls.__members__ = {}

        reserved = set(dir(cls.__class__))
        for key, val in attrs.items():
            if _is_dunder(key):
                continue

            if key in reserved:
                raise ConfigError(f"Attempted to use reserved {key!r} in {name!r}")

            if _is_subclass(val, _Section):
                val.__frozen__ = False
                val.__parent__ = cls
                val.__frozen__ = True

            if _is_config(cls) and _is_section(cls):
                cls.__members__[key] = val
            elif _is_section(cls):
                cls.__members__[key] = val
            elif _is_config(cls):
                if not _is_subclass(val, _Section):
                    raise ConfigError(f"Attempted to add non-Section {key!r} to {cls!r}")
                cls.__members__[key] = val
            else:  # pragma: no cover
                pass

        if _is_config(cls) or _is_section(cls):
            cls.__frozen__ = True

    def __getattribute__(cls, name: str) -> Any:
        attr = super().__getattribute__(name)
        if isinstance(attr, FutureValue):
            return attr.get()
        return attr

    def __getattr__(cls, name: str) -> Any:
        raise ConfigAttrError(f"{name!r} does not exist in {cls!r}")

    def __setattr__(cls, name: str, value: Any) -> None:
        if cls.__frozen__ and name != "__frozen__":
            if name in cls:
                raise ConfigError(f"Attempted to override {name!r} in {cls!r}")
            raise ConfigError(f"Attempted to add {name!r} to {cls!r} at runtime")
        super().__setattr__(name, value)

    def __delattr__(cls, name: str) -> None:
        raise ConfigError(f"Attempted to remove {name!r} from {cls!r}")

    def __getitem__(cls, key: str) -> Any:
        if key in cls:
            return getattr(cls, key)
        raise ConfigKeyError(f"{key!r} does not exist in {cls!r}")

    def __setitem__(cls, key: str, value: Any) -> None:
        return setattr(cls, key, value)

    def __delitem__(cls, key: str) -> None:
        return delattr(cls, key)

    def __call__(cls, *args: Any, **kwargs: Any) -> None:
        if _is_config(cls):
            raise ConfigError(f"Attempted to initialize config {cls!r}")
        elif _is_section(cls):
            raise ConfigError(f"Attempted to initialize section {cls!r}")
        else:  # pragma: nocover
            pass

    @property
    def __members(cls) -> Dict[str, Any]:
        result = {}

        for base in reversed(cls.__bases__):
            if _is_config(base) or _is_section(base):
                for key in base:  # type: ignore
                    result[key] = ...

        for key in cls.__members__:
            result[key] = ...

        return result

    def __len__(cls) -> int:
        return len(cls.__members)

    def __iter__(cls) -> Iterator[str]:
        return cls.__members.__iter__()

    def __contains__(cls, item: Any) -> bool:
        return item in cls.__members

    def __repr__(cls) -> str:
        namespace = [cls.__name__]
        parent = cls.__parent__
        while parent is not None:
            namespace += [parent.__name__]
            parent = parent.__parent__
        return "<" + ".".join(reversed(namespace)) + ">"

    def keys(cls) -> KeysView[str]:
        return cls.__members.keys()

    def values(cls) -> ValuesView[Any]:
        return {key: getattr(cls, key) for key in cls.__members}.values()

    def items(cls) -> ItemsView[str, Any]:
        return {key: getattr(cls, key) for key in cls.__members}.items()

    def get(cls, key: str, default: Union[NilType, Any] = Nil) -> Any:
        try:
            return cls[key]
        except KeyError:
            if default is not Nil:
                return default
            raise

    def __prefetch(cls) -> List[str]:
        errors: List[str] = []
        for key in cls.__members:
            try:
                val = getattr(cls, key)
            except (EnvKeyError, EnvParseError) as e:
                namespace = repr(cls)[1:-1]
                message = f"{namespace}.{key}: {e}"
                errors.append(message)
            else:
                if _is_subclass(val, _Section):
                    errors += val.__prefetch()
        return errors

    def prefetch(cls) -> None:
        errors = cls.__prefetch()
        if len(errors) > 0:
            prefix = "\n- "
            message = f"Failed to prefetch:{prefix}" + prefix.join(errors)
            raise ConfigEnvError(message)

    def __format(cls, *, indent: int = 0, prepend: bool = False) -> List[str]:
        res = [""] if prepend else []

        res.append(" " * indent + f"class <{cls.__name__}>:")
        if len(cls) > 0:
            for key, val in cls.items():
                if _is_subclass(val, (_Config, _Section)):
                    res += val.__format(indent=indent + 4, prepend=True)
                else:
                    res.append(" " * (indent + 4) + f"{key} = {val!r}")
        else:
            res.append(" " * (indent + 4) + "...")

        return res

    def print(cls, stream: Any = sys.stdout) -> None:
        formatted = "\n".join(cls.__format())
        print(formatted, file=stream)


class Section(metaclass=MetaBase):
    __frozen__: bool = False
    __parent__: Optional[MetaBase] = None
    __members__: Dict[str, Any] = {}


_Section = Section


class Config(metaclass=MetaBase):
    __frozen__: bool = False
    __parent__: Optional[MetaBase] = None
    __members__:  Dict[str, Any] = {}


_Config = Config



from ._computed import computed
from ._core import Config, MetaBase, Section
from ._environment import Environment
from ._future_value import FutureValue, ValueType
from ._version import version

__version__ = version
__all__ = ("Config", "Section", "computed",
           "env", "Environment", "FutureValue", "ValueType",
           "MetaBase",)

# type hint for PyCharm
env: Environment = Environment()



from typing import Any, Callable

from ._core import MetaBase


def _required(*args: Any) -> Any:
    pass  # pragma: nocover


class computed:
    def __init__(self, fn: Callable[[Any], Any] = _required) -> None:
        if fn is _required:
            raise TypeError("Use @computed instead of @computed()")
        self._fn = fn

    def __get__(self, _: None, owner: MetaBase) -> Any:
        return self._fn(owner)



import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_int, parse_tuple


@pytest.mark.parametrize(("value", "expected"), [
    ("banana", ("banana",)),
    ("banana,", ("banana",)),
    ("(banana)", ("banana",)),
    ("(banana,)", ("banana",)),
    ("first, second", ("first", "second",)),
    ("first, second,", ("first", "second",)),
    ("(first, second)", ("first", "second",)),
    ("(first, second,)", ("first", "second",)),
])
def test_parse_tuple(value, expected):
    assert parse_tuple(value) == expected


@pytest.mark.parametrize(("value", "expected"), [
    ("banana", ("banana",)),
    ("first second", ("first", "second",))
])
def test_parse_tuple_with_seprator(value, expected):
    assert parse_tuple(value, separator=" ") == expected


@pytest.mark.parametrize(("value", "expected"), [
    ("42", (42,)),
    ("1, 2, 3", (1, 2, 3,))
])
def test_parse_tuple_with_subparser(value, expected):
    assert parse_tuple(value, subparser=parse_int) == expected


def test_parse_empty_tuple():
    with raises(Exception) as exc_info:
        parse_tuple("")

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == ("Failed to parse '' as tuple: "
                                   "Failed to parse '' as non-empty str")


def test_parse_invalid_tuple():
    with raises(Exception) as exc_info:
        parse_tuple("banana,,")

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == ("Failed to parse 'banana,' as tuple: "
                                   "Failed to parse '' as non-empty str")



import sys
from math import isnan

import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_float


@pytest.mark.parametrize(("value", "expected"), [
    ("+3.14", 3.14),
    ("3.14", 3.14),
    ("-3.14", -3.14),
    (".14", 0.14),
    ("3.", 3.0),
    ("0.0", 0.0),
    ("1", 1),
    ("-1", -1),
    ("0", 0),
    ("Infinity", float("inf")),
    ("INFINITY", float("inf")),
    ("inf", float("inf")),
    ("+inf", float("inf")),
    ("-inf", float("-inf")),
    (str(sys.float_info.min), sys.float_info.min),
    (str(sys.float_info.max), sys.float_info.max),
])
def test_parse_float(value, expected):
    assert parse_float(value) == expected


@pytest.mark.parametrize("value", [
    "nan",
    "Nan",
])
def test_parse_float_nan(value):
    assert isnan(parse_float(value))


@pytest.mark.parametrize("value", [
    "",
    ".",
    "3.14.5",
    "3.14a",
    "--3.14",
])
def test_parse_invalid_float(value):
    with raises(Exception) as exc_info:
        parse_float(value)

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == f"Failed to parse {value!r} as float"



import sys

import pytest
from pytest import raises

from cabina.errors import EnvParseError
from cabina.parsers import parse_int


@pytest.mark.parametrize(("value", "expected"), [
    ("1", 1),
    ("+1", 1),
    ("-1", -1),
    ("0", 0),
    (str(sys.maxsize), sys.maxsize),
    (str(-sys.maxsize - 1), -sys.maxsize - 1),
])
def test_parse_int(value, expected):
    assert parse_int(value) == expected


@pytest.mark.parametrize(("value", "base", "expected"), [
    ("00111", 2, 7),
    ("011", 36, 37),
    ("42", 0, 42),
])
def test_parse_int_with_base(value, base, expected):
    assert parse_int(value, base=base) == expected


@pytest.mark.parametrize("value", [
    "",
    "+",
    "12a",
    "--1234",
    "3.14",
])
def test_parse_invalid_int(value):
    with raises(Exception) as exc_info:
        parse_int(value)

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == f"Failed to parse {value!r} as int"


@pytest.mark.parametrize(("value", "base"), [
    ("011", -1),
    ("011", 0),
    ("011", 1),
    ("011", 37),
])
def test_parse_int_with_invalid_base(value, base):
    with raises(Exception) as exc_info:
        parse_int(value, base=base)

    assert exc_info.type is EnvParseError
    assert str(exc_info.value) == f"Failed to parse {value!r} as int with base {base}"



import pytest

from cabina.parsers import parse_as_is


@pytest.mark.parametrize("value", [
    "",
    "None",
    "True",
    "42",
    "3.14",
    "banana",
])
def test_parse_as_is(value):
    assert parse_as_is(value) == value
